// DATQ  -  FIFO utility
// Vn 1.2
// Author - Mike Eggleston
// Copyright 2003 Fenland Software Ltd
// Date 10/06/2003

#define  COPYRIGHT	"Copyright 2003 Fenland Software Ltd"
#define  VERSION	"Vn 1.2"

#include "datq.h"

/* Global Variables */
struct DQ_lock lock_rec;
struct DQ_hdr hdr_rec;
struct DQ_rec dtl_rec;
int DQ_end = 0;
int DQ_catch[] = {CATCHSIGS, 0};

void onsig(int sig)
{
  switch(sig)
  {
    case SIGHUP:
    case SIGALRM: DQ_end = 0;
                  break;
    
    default: DQ_end = 1;
             break;
  }
}

/* Convert a record number into an absolute file position */  
long cvt_recnum(long recnum)
{
long file_pos = (recnum - 1L) * hdr_rec.reclen + sizeof(struct DQ_hdr) + sizeof(struct DQ_lock);
return(file_pos);
}

/* Retrieve the PID of the process monitoring the queue */
int get_monitor_pid(unsigned int fd)
{
struct flock hdr;
  hdr.l_type = F_WRLCK;
  hdr.l_whence = SEEK_SET;
  hdr.l_start = 0L;
  hdr.l_len = sizeof(struct DQ_lock);
  hdr.l_pid= 0;
  if(fcntl(fd,F_GETLK, &hdr) == -1)
    return -1;
  else
    return hdr.l_pid;
}

/* Lock the first few bytes of the file.  Other processes can then 
determine the PID of this process */

int deamon_lock(unsigned int fd, unsigned int lck_type)
{
struct flock hdr;
int monitor_pid;

  /* First check if any other process is monitoring this queue */
  if(monitor_pid = get_monitor_pid(fd))
    return(monitor_pid);
  else
  {
    /* OK - no other process has this queue so we can flag it for 
       this process */

    hdr.l_type = lck_type;
    hdr.l_whence = SEEK_SET;
    hdr.l_start = 0L;
    hdr.l_len = sizeof(struct DQ_lock);
    hdr.l_pid = 0;

    if(fcntl(fd,F_SETLK, &hdr) == -1)
      return(-1);
    else
      return 0;
  }
}

/* Lock the header record for updating.
   If it's already locked, wait for it to be released */
int lock_header(unsigned int fd, unsigned int lck_type)
{
struct flock hdr;
  hdr.l_type = lck_type;
  hdr.l_whence = SEEK_SET;
  hdr.l_start = (long)sizeof(struct DQ_lock);
  hdr.l_len = sizeof(struct DQ_hdr);
  if(fcntl(fd,F_SETLKW, &hdr) == -1)
    return(errno);
  else
    return 0;
}

int read_header(unsigned int fd)
{
int rslt;

  if((rslt = lock_header(fd,F_WRLCK)) == 0)
  {
    lseek(fd, (long)sizeof(struct DQ_lock), SEEK_SET);
    rslt = (read(fd, &hdr_rec, sizeof(struct DQ_hdr))) - sizeof(struct DQ_hdr);
  }

  if(rslt)
    return(errno);
  else
    return 0;
}

int create_queue(char *qname, int reclen)
{
unsigned int fd;
int error_flag = 0;

  lock_rec.monitor_pid = 0;
  hdr_rec.reclen = sizeof(struct DQ_rec) + reclen;
  hdr_rec.dtalen = reclen;
  hdr_rec.last_wrtn = 0L;
  hdr_rec.next_read = 0L;
  hdr_rec.last_delt = 0L;
  hdr_rec.queue_len = 0;

  if((fd = open(qname, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR|S_IWUSR)) == -1)
  {
    fprintf(stderr,"Create Failed - %s\n", strerror(errno));
    return(errno);
  }
  else
  {
    if(write(fd, &lock_rec, sizeof(struct DQ_lock)) != sizeof(struct DQ_lock))
      error_flag = 1;
    else
    {
      if(write(fd, &hdr_rec, sizeof hdr_rec) != sizeof hdr_rec)
        error_flag = 1; 
      else
      {
        close(fd);
        fprintf(stderr,"Dataqueue %s created.\n", qname);
      }
    }
  }

  if(error_flag)
  {
    fprintf(stderr,"Write of Header Failed - %s\n", strerror(errno));
    close(fd);
    unlink(qname);
    return(errno);
  }
return 0;
}

int add_record(char *qname, int add_to_top)
{
int rtn_cde=0;
unsigned int fd;
long recnum;
char *rec;
pid_t monitor;

  if((fd = open(qname, O_RDWR)) == -1)
  {
    fprintf(stderr,"Dataqueue %s not found\n", qname);
    rtn_cde= -1;
  }
  else
  {
    if(read_header(fd))
    {
      fprintf(stderr,"Add Failed - %s\n", strerror(errno));
      close(fd);
      return(errno);
    }

    rec = (char *)malloc(hdr_rec.dtalen);
    fgets(rec,hdr_rec.dtalen,stdin);

    if(hdr_rec.last_wrtn == 0L)
    {
      recnum = lseek(fd, 0L, SEEK_END);
      hdr_rec.last_wrtn = ((recnum - sizeof(struct DQ_hdr)) / 
hdr_rec.reclen) + 1L;
      dtl_rec.prev_rec = 0L;
      dtl_rec.next_rec = 0L;
      write(fd, &dtl_rec, sizeof(struct DQ_rec));
      write(fd, rec, hdr_rec.dtalen);
      hdr_rec.queue_len++;
      hdr_rec.next_read = 1L;
      lseek(fd, sizeof(struct DQ_lock), SEEK_SET);
      write(fd, &hdr_rec, sizeof(struct DQ_hdr));
      close(fd);
    }
    else
    {
      if(hdr_rec.last_delt == 0L)
      {
        recnum = lseek(fd, 0L, SEEK_END);
        if(add_to_top)
        {
          dtl_rec.prev_rec = 0L;
          dtl_rec.next_rec = hdr_rec.next_read;
          hdr_rec.next_read = ((recnum - sizeof(struct DQ_hdr)) / hdr_rec.reclen) + 1L;
        }
        else
        {
          dtl_rec.prev_rec = hdr_rec.last_wrtn;
          dtl_rec.next_rec = 0L;
          hdr_rec.last_wrtn = ((recnum - sizeof(struct DQ_hdr)) / hdr_rec.reclen) + 1L;
        }
        write(fd, &dtl_rec, sizeof(struct DQ_rec));
        write(fd, rec, hdr_rec.dtalen);
        if(add_to_top)
        {
          recnum = lseek(fd, cvt_recnum(dtl_rec.next_rec), SEEK_SET);
          read(fd, &dtl_rec, sizeof(struct DQ_rec));
          dtl_rec.prev_rec = hdr_rec.next_read;
        }
        else
        {
          recnum = lseek(fd, cvt_recnum(dtl_rec.prev_rec), SEEK_SET);
          read(fd, &dtl_rec, sizeof(struct DQ_rec));
          dtl_rec.next_rec = hdr_rec.last_wrtn;
        }
        lseek(fd, recnum, SEEK_SET);
        write(fd, &dtl_rec, sizeof(struct DQ_rec));
        hdr_rec.queue_len++;
        if(hdr_rec.next_read == 0L)
          hdr_rec.next_read = hdr_rec.last_wrtn;
        lseek(fd, sizeof(struct DQ_lock), SEEK_SET);
        write(fd, &hdr_rec, sizeof(struct DQ_hdr));
        close(fd);
      }
      else
      {
        recnum = lseek(fd, cvt_recnum(hdr_rec.last_delt), SEEK_SET);
        read(fd, &dtl_rec, sizeof(struct DQ_rec));
        hdr_rec.last_delt = dtl_rec.prev_rec;
        if(add_to_top)
        {
          dtl_rec.prev_rec = 0L;
          dtl_rec.next_rec = hdr_rec.next_read;
          hdr_rec.next_read = ((recnum - sizeof(struct DQ_hdr)) / hdr_rec.reclen) + 1L;
        }
        else
        {
          if(hdr_rec.queue_len)
            dtl_rec.prev_rec = hdr_rec.last_wrtn;
          else
            dtl_rec.prev_rec = 0L;

          dtl_rec.next_rec = 0L;
          hdr_rec.last_wrtn = ((recnum - sizeof(struct DQ_hdr)) / hdr_rec.reclen) + 1L;
        }

        lseek(fd, recnum, SEEK_SET);
        write(fd, &dtl_rec, sizeof(struct DQ_rec));
        write(fd, rec, hdr_rec.dtalen);
        if(add_to_top)
        {
          recnum = lseek(fd, cvt_recnum(dtl_rec.next_rec), SEEK_SET);
          read(fd, &dtl_rec, sizeof(struct DQ_rec));
          dtl_rec.prev_rec = hdr_rec.next_read;
        }
        else
        {
          recnum = lseek(fd, cvt_recnum(dtl_rec.prev_rec), SEEK_SET);
          read(fd, &dtl_rec, sizeof(struct DQ_rec));
          dtl_rec.next_rec = hdr_rec.last_wrtn;
        }
        lseek(fd, recnum, SEEK_SET);
        write(fd, &dtl_rec, sizeof(struct DQ_rec));
        hdr_rec.queue_len++;
        if(hdr_rec.next_read == 0L)
          hdr_rec.next_read = hdr_rec.last_wrtn;
        lseek(fd, sizeof(struct DQ_lock), SEEK_SET);
        write(fd, &hdr_rec, sizeof(struct DQ_hdr));
      }
    }

    /* Wake up the monitoring process (if any) */
    if((monitor = get_monitor_pid(fd)) > 0)
      kill(monitor, SIGHUP);

    close(fd);
  }
return(rtn_cde);
}

int read_record(char *qname, int waitq)
{
int rtn_cde=0;
unsigned int fd;
unsigned long recnum;
unsigned long curr_rec;
char *rec;
pid_t current_pid = getpid();
pid_t monitor_pid = 0;

  if((fd = open(qname, O_RDWR)) == -1)
  {
    fprintf(stderr,"Dataqueue %s not found\n", qname);
    rtn_cde= -1;
  }
  else
  {
    if(monitor_pid = deamon_lock(fd, F_WRLCK))
    {
      fprintf(stderr,"Data Queue already assigned to PID %d\n", 
               monitor_pid);
      close(fd);
      return -1;
    }

    deamon_lock(fd, F_WRLCK);

    do
    {
      if(read_header(fd))
      {
        fprintf(stderr,"Read Failed - %s\n", strerror(errno));
        close(fd);
        return(errno);
      }

      if(waitq && (hdr_rec.next_read == 0L))
      {
        lock_header(fd,F_UNLCK);
        pause();
        if(DQ_end)
          break;
        else
          continue; 
      }

      rec=(char *)malloc(hdr_rec.dtalen);

      if(hdr_rec.next_read)
      {
        /* Read the 'detail' record and send it to stdout */ 
        curr_rec = hdr_rec.next_read;
        lseek(fd, cvt_recnum(curr_rec), SEEK_SET);
        read(fd, &dtl_rec, sizeof(struct DQ_rec));
        read(fd, rec, hdr_rec.dtalen);
        hdr_rec.next_read = dtl_rec.next_rec;
        dtl_rec.prev_rec = hdr_rec.last_delt;
        dtl_rec.next_rec = 0L;
        hdr_rec.last_delt = curr_rec;
        lseek(fd, cvt_recnum(curr_rec), SEEK_SET);
        write(fd, &dtl_rec, sizeof(struct DQ_rec));
        hdr_rec.queue_len--;
        lseek(fd, sizeof(struct DQ_lock), SEEK_SET);
        write(fd, &hdr_rec, sizeof(struct DQ_hdr));
        printf("%s", rec);
        fflush(stdout);
      }
      lock_header(fd, F_UNLCK);
    } while(waitq);

    /* Clear monitor lock */
    deamon_lock(fd, F_UNLCK);

    close(fd);
  }
return(rtn_cde);
}

int usage()
{
  printf("datq Vn 1.2\n");
  printf("Usage Instructions\n");
  printf("Options -\n -c Create\n -a/-t Add a record\n -r Read a record\n -w Wait for records\n");
  printf(" -v Verbose mode (for debugging)\n\n");
  printf("Examples -\n");
  printf("datq -c 60 queue  /* Create a queue called 'queue' with records of 60 chars*/\n");
  printf("datq -a queue     /* Add a record to 'queue' */\n");
  printf("datq -t queue     /* Add a record to begining of 'queue' */\n");
  printf("datq -r queue     /* Read a record from 'queue' */\n");
  printf("datq -w queue     /* Monitor 'queue' and read records as they appear */\n");
return 0;
}
 
int main(int argc, char *argv[])
{
char *qname;
int i, opt, action = 'h', dtalen=0;
int rtn_cde;

  for(i=0; DQ_catch[i]; i++)
    if(signal(DQ_catch[i], onsig) == SIG_ERR)
      fprintf(stderr, "Can't set signal %d handler:", DQ_catch[i]);
 
  while((opt = getopt(argc,argv,"c:arhtvw")) != -1)
  {
    switch(opt)
    {
      case 'c':  if((dtalen = atoi(optarg)) == 0) {
                   fprintf(stderr, "Invalid record size");
        	   action = 'E';
                   break;
                 }

      case 'a':
      case 't':
      case 'r':
      case 'w':  if(action == 'h')
                   action = opt;
                 else
                 {
                   action = 'E';
                   opt = -1;
                 }  
                 break;

      default:   action = 'E';
                 opt = -1;
                 break;
    }
  }

  if(optind < argc)
  {
    qname=(char *)malloc(strlen(argv[optind])*sizeof(char));
    strcpy(qname,argv[optind]);
  }
  else 
    action = 'E';

  if(qname && (action != 'E'))
  {
    switch(action)
    {
      case 'c':  rtn_cde = create_queue(qname, dtalen);
                 break;
      case 'a':  rtn_cde = add_record(qname, 0);
                 break;
      case 't':  rtn_cde = add_record(qname, 1);
                 break;
      case 'r':  rtn_cde = read_record(qname, 0);
                 break;
      case 'w':  rtn_cde = read_record(qname, 1);
                 break;
    }
  }
  else
    usage();

exit(rtn_cde);
}  
